@@ -76,6 +76,8 @@
                      sock :: gen_udp:socket(),
                      loc_addr :: inet:ip_address(),
                      rem_addr :: inet:ip_address(),
+                     loc_rts :: pos_integer(),
+                     rem_rts :: undefined | pos_integer(),
                      seq_nr :: pfcp_seq_nr(),
                      registry :: dict:dict()
                     }).
@@ -142,6 +144,7 @@ init([LocAddr, RemAddr]) ->
                                  loc_addr = LocAddr,
                                  rem_addr = RemAddr,
                                  seq_nr = 0,
+                                 loc_rts = get_recovery_timestamp(),
                                  registry = dict:new()}}.
 
 
@@ -164,18 +167,27 @@ connecting(state_timeout, assoc_setup_timeout, S) ->
     {repeat_state, S};
 
 %% Handle incoming PFCP PDU(s)
-connecting(info, {udp, Sock, _FromIp, _FromPort, Data},
+connecting(info, {udp, Sock, FromIp, FromPort, Data},
            #peer_state{sock = Sock} = S) ->
-    case decode_pdu(Data, S) of
+    PDU = decode_pdu(Data, {FromIp, FromPort}, S),
+    case PDU of
         #pfcp{type = association_setup_response,
-              ie = #{pfcp_cause := 'Request accepted'}} ->
+              ie = #{pfcp_cause := 'Request accepted',
+                     recovery_time_stamp := #recovery_time_stamp{time = RRTS}}} ->
             ?LOG_INFO("Rx Association Setup Response (Request accepted)"),
-            {next_state, connected, S};
+            {next_state, connected,
+             S#peer_state{rem_rts = RRTS}};
         #pfcp{type = association_setup_response,
               ie = #{pfcp_cause := Cause}} ->
             ?LOG_ERROR("Rx Association Setup Response (~p)", [Cause]),
             {stop, {shutdown, assoc_setup_nack}};
-        PDU ->
+        %% 3GPP TS 29.244, 6.2.2.2 Heartbeat Request
+        %% A CP function or UP function shall be prepared to receive a Heartbeat Request
+        %% at any time (even from unknown peers) and it shall reply with a Heartbeat Response.
+        #pfcp{type = heartbeat_request} ->
+            recv_heartbeat_request(PDU, {FromIp, FromPort}, S),
+            {keep_state, S};
+        _ ->
             ?LOG_NOTICE("Rx unexpected PFCP PDU: ~p", [PDU]),
             {keep_state, S}
     end;
@@ -208,10 +220,16 @@ connected({call, From},
     {keep_state, S1, [{reply, From, Result}]};
 
 %% Handle incoming PFCP PDU(s)
-connected(info, {udp, Sock, _FromIp, _FromPort, Data},
+connected(info, {udp, Sock, FromIp, FromPort, Data},
           #peer_state{sock = Sock} = S) ->
-    PDU = decode_pdu(Data, S),
-    route_pdu(PDU, S),
+    PDU = decode_pdu(Data, {FromIp, FromPort}, S),
+    case PDU of
+        %% 3GPP TS 29.244, 6.2.2.2 Heartbeat Request
+        #pfcp{type = heartbeat_request} ->
+            recv_heartbeat_request(PDU, {FromIp, FromPort}, S);
+        _ ->
+            route_pdu(PDU, S)
+    end,
     {keep_state, S};
 
 %% Catch-all handler for this state
@@ -282,11 +300,13 @@ registry_del({Pid, Ref}, #peer_state{registry = Reg} = S) ->
     S#peer_state{registry = NReg}.
 
 
--spec decode_pdu(binary(), peer_state()) -> pfcp_pdu().
-decode_pdu(Data, _S) ->
-    ?LOG_DEBUG("Rx encoded PFCP PDU: ~p", [Data]),
+-spec decode_pdu(binary(), tuple(), peer_state()) -> pfcp_pdu().
+decode_pdu(Data, {FromIp, FromPort}, _S) ->
+    ?LOG_DEBUG("Rx encoded PFCP PDU from ~p:~p: ~p",
+               [FromIp, FromPort, Data]),
     PDU = pfcp_packet:decode(Data),
-    ?LOG_DEBUG("Rx PFCP PDU: ~p", [PDU]),
+    ?LOG_DEBUG("Rx PFCP PDU from ~p:~p: ~p",
+               [FromIp, FromPort, PDU]),
     %% TODO: check PDU#pfcp.seq_no
     %% TODO: there can be batched PDUs
     PDU.
@@ -307,7 +327,8 @@ route_pdu(#pfcp{seid = SEID} = PDU,
 -spec send_data(binary(), peer_state()) -> ok | {error, term()}.
 send_data(Data, #peer_state{sock = Sock,
                             rem_addr = RemAddr}) ->
-    ?LOG_DEBUG("Tx encoded PFCP PDU: ~p", [Data]),
+    ?LOG_DEBUG("Tx encoded PFCP PDU to ~p:~p: ~p",
+               [RemAddr, ?PFCP_PORT, Data]),
     gen_udp:send(Sock, RemAddr, ?PFCP_PORT, Data).
 
 
@@ -365,9 +386,9 @@ get_f_seid(SEID, S) ->
 
 %% 6.2.6 PFCP Association Setup Procedure
 %% 7.4.4.1 PFCP Association Setup Request
-send_assoc_setup(S) ->
+send_assoc_setup(#peer_state{loc_rts = LRTS} = S) ->
     IEs = #{node_id => #node_id{id = get_node_id(S)},
-            recovery_time_stamp => #recovery_time_stamp{time = get_recovery_timestamp()}},
+            recovery_time_stamp => #recovery_time_stamp{time = LRTS}},
     send_pdu({association_setup_request, IEs}, S).
 
 
@@ -401,4 +422,20 @@ send_session_modify_req(SEID, PDRs, FARs, S) ->
 send_session_delete_req(SEID, S) ->
     send_pdu(SEID, {session_deletion_request, []}, S).
 
+
+%% 6.2.2 Heartbeat Procedure
+%% 7.4.2 Heartbeat Messages
+recv_heartbeat_request(#pfcp{type = heartbeat_request,
+                             seq_no = SeqNr,
+                             ie = ReqIEs},
+                       {FromIp, FromPort},
+                       #peer_state{loc_rts = LRTS} = S) ->
+    ?LOG_INFO("Rx Heartbeat Request from ~p:~p: ~p", [FromIp, FromPort, ReqIEs]),
+    RspIEs = #{recovery_time_stamp => #recovery_time_stamp{time = LRTS}},
+    %% Fake #peer_state to force the right destination and SeqNr
+    {Result, _} = send_pdu({heartbeat_response, RspIEs},
+                           S#peer_state{seq_nr = SeqNr,
+                                        rem_addr = FromIp}),
+    Result.
+
 %% vim:set ts=4 sw=4 et:
